import os
from docopt import docopt, DocoptExit
try:
    import logger as log
except ImportError:
    from jsc import logger as log
import shlex


class RecipeRuntimeError(Exception):
    pass


def docopt_cmd(func):
    """
    This decorator is used to simplify the try/except block and pass the result
    of the docopt parsing to the called action.
    """
    def fn(remoto_exec, state, arg):
        # make sure it's of type str and not unicode/byte
        # shlex tokenizes the string like python does for sys.argv
        # arg = shlex.split("{}".format(arg))
        log.info(arg)
        arg = shlex.split(arg)
        opt = docopt(func.__doc__, arg)
        return func(remoto_exec, state, opt)
    return fn


@docopt_cmd
def rc_name(remoto_exec, state, args):
    """
    Usage:
      name <name>
    """
    state["name"] = args["<name>"]
    return state


@docopt_cmd
def rc_package(remoto_exec, state, args):
    """
    Usage:
      package <pkg>...
    """
    success, installed_packages = remoto_exec.package(args["<pkg>"])
    if not success:
        raise RecipeRuntimeError("package failed")
    if "package" not in state["software_list"]:
        state["software_list"]["package"] = {}
    for pkg in installed_packages.keys():
        state["software_list"]["package"][pkg] = {}
        state["software_list"]["package"][pkg]["version"] = installed_packages[pkg]
    return state


@docopt_cmd
def rc_gd(remoto_exec, state, args):
    """
    Usage:
      gd [--pkey=<pkey>] [--branch=<branch>] [--depth=<n>] <repo> <dst>
    """
    is_dev = state["is_dev"]
    repo = args["<repo>"]
    dst = args["<dst>"]
    depth = args["--depth"] if is_dev else 1
    branch = args["--branch"]
    pkey = args["--pkey"]
    success, git_res = remoto_exec.git_clone(repo, dst, depth, branch, pkey)
    commit, ref = git_res
    if success:
        if "gd" not in state["software_list"]:
            state["software_list"]["gd"] = {}
        state["software_list"]["gd"][args["<dst>"]] = {
            "ref": ref,
            "commit": commit,
            "repo": args["<repo>"]
            }
        if not is_dev:
            remoto_exec.rmtree(os.path.join(args["<dst>"], ".git"))
        return state
    raise RecipeRuntimeError("gd failed")


# no docopt_cmd since we want to forward the full argument line to remoto.rc_run
def rc_run(remoto_exec, state, cmd):
    """
    Usage:
      run <cmd>...
    """
    success, msg = remoto_exec.rc_run(cmd)
    if not success:
        raise RecipeRuntimeError(msg)
    else:
        log.white(msg)
    return state


@docopt_cmd
def rc_install(remoto_exec, state, args):
    """
    Usage:
      install <src> <dst>
    """
    success, msg = remoto_exec.install(args['<src>'], args['<dst>'])
    if not success:
        raise RecipeRuntimeError(msg)
    return state


@docopt_cmd
def rc_append(remoto_exec, state, args):
    """
    Usage:
      append <file> <text>
    """
    success, msg = remoto_exec.file_put(args['<file>'], args['<text>'])
    if not success:
        raise RecipeRuntimeError(msg)
    return state


@docopt_cmd
def rc_put(remoto_exec, state, args):
    """
    Usage:
      put <file> <text>
    """
    success, msg = remoto_exec.file_put(args['<file>'], args['<text>'], truncate=True)
    if not success:
        raise RecipeRuntimeError(msg)
    return state


@docopt_cmd
def rc_replace(remoto_exec, state, args):
    """
    Usage:
      replace <file> <find> <replace>
    """
    success, msg = remoto_exec.file_content_replace(args['<file>'], args['<find>'], args['<replace>'])
    if not success:
        raise RecipeRuntimeError(msg)
    return state


@docopt_cmd
def rc_insert(remoto_exec, state, args):
    """
    Usage:
      insert <file> <find> <insert>
    """
    success, msg = remoto_exec.file_content_replace(args['<file>'], args['<find>'], args['<find>']+args['<insert>'], accurances=1)
    if not success:
        raise RecipeRuntimeError(msg)
    return state


@docopt_cmd
def rc_rinsert(remoto_exec, state, args):
    """
    Usage:
      rinsert <file> <find> <insert>
    """
    success, msg = remoto_exec.file_content_replace(args['<file>'], args['<find>'], args['<insert>']+args['<find>'], accurances=1, reverse=True)
    if not success:
        raise RecipeRuntimeError(msg)
    return state


cmd_map = {
    "name": rc_name,
    "package": rc_package,
    "gd": rc_gd,
    "run": rc_run,
    "install": rc_install,
    "append": rc_append,
    "put": rc_put,
    "replace": rc_replace,
    "insert": rc_insert,
    "rinsert": rc_rinsert}


def recipe_execute_cmd(remoto_exec, state, line):
    line_words = line.split(' ')
    command_word = line_words[0]
    args = ' '.join(line_words[1:])
    state = cmd_map[command_word](remoto_exec, state, args)


def run(remoto_exec, recipe, is_dev):
    lc = 1
    state = {"software_list": {}, "name": None, "is_dev": is_dev}
    try:
        for line in recipe.splitlines(False):
            # Everything after # is comment.
            clean_line = line.strip().split("#")[0]
            if clean_line == '':
                continue
            recipe_execute_cmd(remoto_exec, state, clean_line)
            lc += 1
        return state
    except DocoptExit as e:
        log.err('Recipe contains an invalid command at Line {line}.'.format(line=lc))
        log.err(e)
    except RecipeRuntimeError as e:
        log.err('Recipe command on line {line} failed with error: {e}'.format(line=lc, e=e))



__version__ = "1.0.3"
__description__ = "Jumpstarter Console"
__long_description__ = """
Jumpstarter Console - A command line tool for managing your Jumpstarter projects.
please visit https://github.com/jumpstarter-io/jsc for more information
"""



import datetime
import fcntl
import os
import os.path
import json
import re
import select
import subprocess
import sys
import threading
import shutil

try:
    import urllib2 as url
    from urllib2 import HTTPError

except ImportError:
    import urllib.request as url
    from urllib.error import HTTPError

try:
    import httplib
except ImportError:
    import http.client as httplib

"""
This module is indended to be executed on remote side.
Although it's designed to run in python3, it still needs support for python2 since it's imported as a module locally
"""


CODE_DIR = "/app/code"
STATE_DIR = "/app/state"
JSC_DIR = os.path.join(CODE_DIR, ".jsc")
BACKUPS_DIR = os.path.join(JSC_DIR, "backups")
BACKUPS_SEQ_FILE_PATH = os.path.join(BACKUPS_DIR, "seq")
LOCK_FILE = os.path.join(JSC_DIR, "lock")
RECIPE_PATH = os.path.join(JSC_DIR, "recipe")
NEW_RECIPE_PATH = os.path.join(JSC_DIR, "new-recipe")
NEW_RECIPE_SRC = os.path.join(NEW_RECIPE_PATH, "src")
NEW_RECIPE_SCRIPT = os.path.join(NEW_RECIPE_SRC, "Jumpstart-Recipe")


RECIPE_PATH = os.path.join(JSC_DIR, "recipe")

class bcolors:
    HEADER = '\033[95m'
    OKBLUE = '\033[94m'
    OKGREEN = '\033[92m'
    WARNING = '\033[93m'
    FAIL = '\033[91m'
    ENDC = '\033[0m'
    BOLD = "\033[1m"

    def disable(self):
        self.HEADER = ''
        self.OKBLUE = ''
        self.OKGREEN = ''
        self.WARNING = ''
        self.FAIL = ''
        self.ENDC = ''


print_lock = threading.Lock()


def print_locked(message, bcolor, printstr="%s%s%s\n"):
    with print_lock:
        sys.stderr.write(printstr % (bcolor, message, bcolors.ENDC))
        sys.stderr.flush()


def log(message):
    print_locked(message, bcolors.HEADER)


def log_ok(message):
    print_locked(message, bcolors.OKGREEN)


def log_warn(message):
    print_locked(message, bcolors.WARNING)


def fail(message):
    print_locked(message, bcolors.FAIL, "%sError: %s%s\n")
    os._exit(1)


def pythonify_args(func):
    """
    Used as a decorator to convert arguments from remoto into pythonesque equivalents
    """
    def fn(*args, **kwargs):
        if len(args) > 0 and args[0] == "remoto":
            arg_list = args[1]
            return func(*arg_list[0], **arg_list[1])
        else:
            return func(*args, **kwargs)
    return fn


class AssemblyStateError(Exception):
    """Base class for exceptions in this module."""
    pass


def touch_dir(directory):
    if not os.path.exists(directory):
        try:
            os.makedirs(directory)
        except FileExistsError:
            pass


def adjust_remote_pwd(file_path):
    if not file_path.startswith("/"):
        return os.path.expanduser(os.path.join(NEW_RECIPE_SRC, file_path))
    return file_path


@pythonify_args
def file_put(file_path, content, truncate=False):
    file_path = adjust_remote_pwd(file_path)
    if not os.path.isdir(os.path.dirname(file_path)):
        return False, "Path does not exist"
    open_flag = "a+"
    if truncate is not False:
        open_flag = "w+"
    with open(file_path, open_flag) as f:
        if truncate:
            f.truncate(0)
        f.write(content)
    return True, None


@pythonify_args
def file_content_replace(file_path, find, replace, accurances=None, reverse=False):
    file_path = adjust_remote_pwd(file_path)
    try:
        with open(file_path, "r") as f:
            content = f.read()
        with open(file_path, "w") as f:
            f.truncate(0)
            if accurances is None:
                accurances = len(content)
            if not reverse:
                f.write(content.replace(find, replace, accurances))
            else:
                f.write(content[::-1].replace(find[::-1], replace[::-1], accurances)[::-1])
        return True, None
    except (FileNotFoundError, IOError):
        return False, "file {file_path} not found".format(file_path=file_path)


@pythonify_args
def install(src, dst):
    src = adjust_remote_pwd(src)
    try:
        if os.path.isdir(src):
            shutil.copytree(src, dst, symlinks=True)
        else:
            shutil.copy2(src, dst)
    except (FileNotFoundError, IOError) as e:
        return False, str(e)
    return True, None


@pythonify_args
def rc_run(cmd):
    wd = os.getcwd()
    os.chdir(NEW_RECIPE_SRC)
    success = True
    try:
        output = subprocess.check_output(cmd, shell=True).decode("utf-8")
    except subprocess.CalledProcessError as e:
        output = str(e)
    os.chdir(wd)
    return success, output


@pythonify_args
def env_json():
    with open("/app/env.json") as f:
        container_env = json.loads(f.read())
        return container_env


@pythonify_args
def assert_is_assembly():
    if not env_json()['ident']['container']['is_assembly']:
        return False, "You tried to connect to a non-assembly container"
    return True, None


@pythonify_args
def assert_assembly_is_empty():
    for node in os.listdir(CODE_DIR):
        if node != 'lost+found':
            return False, "{code_dir} is not empty, cannot continue".format(code_dir=CODE_DIR)
    return True, None


@pythonify_args
def lock_session(lock_content):
    fd = os.open(LOCK_FILE, os.O_CREAT)
    try:
        fcntl.flock(fd, fcntl.LOCK_EX | fcntl.LOCK_NB)
        with open(LOCK_FILE, "w") as f:
            f.truncate(0)
            f.write(lock_content)
            f.flush()
        return True, None
    except IOError:
        with open(LOCK_FILE) as f:
            lock_content = json.loads(f.read())
        return False, "File lock is already acquired by [{who}] since [{when}]".format(who=lock_content['hostname'], when=lock_content['unix_epoch'])


@pythonify_args
def check_init():
    for node in (JSC_DIR, BACKUPS_DIR):
        if not os.path.exists(node):
            return False, None
    if not os.path.exists(BACKUPS_SEQ_FILE_PATH):
        return False, None
    return True, None


@pythonify_args
def init():
    for node in (JSC_DIR, BACKUPS_DIR):
        if not os.path.exists(node):
            touch_dir(node)
    if not os.path.exists(BACKUPS_SEQ_FILE_PATH):
        with open(BACKUPS_SEQ_FILE_PATH, 'w+') as f:
            f.write("1")
    if os.path.exists(NEW_RECIPE_PATH):
        shutil.rmtree(NEW_RECIPE_PATH)


@pythonify_args
def sync_dir(directory):
    fd = os.open(directory, os.O_DIRECTORY)
    os.fsync(fd)
    os.close(fd)


@pythonify_args
def status_short(local_version, verbose):
    env = env_json()
    container_id = env["ident"]["container"]["id"]
    email = env["ident"]["user"]["email"]
    name = env["ident"]["user"]["name"]
    info = {"container_id": container_id, "email": email, "name": name, "version": local_version}
    log_ok("jsc v{version} attached to assembly [{container_id}] by [{name} <{email}>]".format(**info))
    if verbose:
        try:
            with open(os.path.join(RECIPE_PATH, "software-list")) as f:
                software = json.loads(f.read())
                package_lines = "\n".join(["\t{pkg}: [{ver}]".format(pkg=pkg, ver=software["package"][pkg]["version"]) for pkg in software["package"]])
                gd_lines = "\n".join(["\t{path}: [{ref}] [{src}] [{commit}]".format(path=path,
                                                                                    ref=software["gd"][path]["ref"],
                                                                                    src=software["gd"][path]["repo"],
                                                                                    commit=software["gd"][path]["commit"][0:8])
                                      for path in software["gd"]])
                log_ok("\nDeployed packages:\n{pkg}\n\nGit deployed software:\n{gd}".format(pkg=package_lines, gd=gd_lines))
        except FileNotFoundError:
            pass
    return True


@pythonify_args
def is_code_dir_clean():
    for node in os.listdir(CODE_DIR):
        if node not in ("lost+found", ".jsc", ".pacman", ".config"):
            return False
    return True


@pythonify_args
def backup_new():
    log("Backup starting")
    recipe_dir = os.path.join(JSC_DIR, "recipe")
    new_backup_dir = os.path.join(BACKUPS_DIR, "new-backup")
    size_file = os.path.join(new_backup_dir, "size")
    touch_dir(new_backup_dir)
    with open(BACKUPS_SEQ_FILE_PATH) as f:
        seq_num = int(f.read())
    with open(BACKUPS_SEQ_FILE_PATH, "w") as f:
        f.truncate(0)
        f.write(str(seq_num + 1))
        f.flush()
    log("Compressing")
    new_backup_file = os.path.join(new_backup_dir, "data.tar.lzo")
    pacman_dir = os.path.join(CODE_DIR, ".pacman")
    if not os.path.isdir(pacman_dir):
        pacman_dir = ""
    subprocess.check_call("tar --use-compress-program=lzop --exclude='{code_dir}/.pacman/cache' --exclude='{code_dir}/.pacman/db/sync' --exclude='lost+found' -cf {new_backup_file} {code_dir}/* {pacman_dir}".format(new_backup_file=new_backup_file, code_dir=CODE_DIR, pacman_dir=pacman_dir), shell=True)
    lzop_info = subprocess.check_output("lzop --info {new_backup_file}".format(new_backup_file=new_backup_file).split(" ")).decode("utf-8")
    for entry in lzop_info.split(" "):
        # the first digit we find is the uncompressed size
        if entry.isdigit():
            log("Uncompressed size {}".format(lzop_info))
            with open(size_file, "w+") as f:
                f.truncate(0)
                f.write(entry)
                f.flush
            break
    log("Saving recipe")
    if os.path.isdir(recipe_dir):
        subprocess.check_call("cp -r {recipe_dir} {new_backup_dir}".format(recipe_dir=recipe_dir, new_backup_dir=new_backup_dir), shell=True)
    sync_dir(CODE_DIR)
    created_time = datetime.datetime.utcnow().strftime("%Y-%m-%dT%H:%M:%SUTC")
    backup_target_dir = os.path.join(BACKUPS_DIR, "{seq_num}@{created_time}".format(seq_num=seq_num, created_time=created_time))
    log("Finnishing up")
    subprocess.check_call("mv {new_backup_dir} {backup_target_dir}".format(new_backup_dir=new_backup_dir, backup_target_dir=backup_target_dir), shell=True)


re_backup_name = re.compile(r"(\d)+@(\d{4}-\d{2}-\d{2})T(\d{2}:\d{2}:\d{2})([\w\-\+]+)")


@pythonify_args
def backup_ls():
    backup_list = []
    for node in os.listdir(BACKUPS_DIR):
        match = re.match(re_backup_name, node)
        if match is not None:
            m_id = match.group(1)
            m_date = match.group(2)
            m_time = match.group(3)
            m_tz = match.group(4)
            backup_dir = os.path.join(BACKUPS_DIR, node)
            jumpstart_recipe_path = os.path.join(backup_dir, "recipe", "src", "Jumpstart-Recipe")
            if os.path.isfile(jumpstart_recipe_path):
                with open(jumpstart_recipe_path) as f:
                    for line in f.readlines():
                        if line.startswith("name"):
                            m_recipe_name = line.split(" ")[1].strip()
                            break
            else:
                m_recipe_name = "<broken/unknown>"
            backup_list.append("{id}: {date} {time} {tz}, {recipe_name}".format(id=m_id, date=m_date, time=m_time, tz=m_tz, recipe_name=m_recipe_name))
    return backup_list


def sizeof_fmt(num, suffix='B'):
    for unit in ['', 'Ki', 'Mi', 'Gi', 'Ti', 'Pi', 'Ei', 'Zi']:
        if abs(num) < 1024.0:
            return "%3.1f%s%s" % (num, unit, suffix)
        num /= 1024.0
    return "%.1f%s%s" % (num, 'Yi', suffix)


@pythonify_args
def backup_du():
    # STEP1: total disk size
    stat_code_dir = os.statvfs(CODE_DIR)
    code_dir_total_b = stat_code_dir.f_frsize * stat_code_dir.f_blocks
    backup_list = []
    for node in os.listdir(BACKUPS_DIR):
        match = re.match(re_backup_name, node)
        if match is not None:
            m_id = match.group(1)
            m_date = match.group(2)
            m_time = match.group(3)
            m_tz = match.group(4)
            backup_dir = os.path.join(BACKUPS_DIR, node)
            jumpstart_recipe_path = os.path.join(backup_dir, "recipe", "src", "Jumpstart-Recipe")
            if os.path.isfile(jumpstart_recipe_path):
                with open(jumpstart_recipe_path) as f:
                    for line in f.readlines():
                        if line.startswith("name"):
                            m_recipe_name = line.split(" ")[1].strip()
                            break
            else:
                m_recipe_name = "<broken/unknown>"
            backup_file = os.path.join(backup_dir, "data.tar.lzo")
            backup_file_stat = os.stat(backup_file)
            percent_of_total = "{:.1%}".format(backup_file_stat.st_size / code_dir_total_b)
            backup_file_size = sizeof_fmt(backup_file_stat.st_size)
            with open(os.path.join(backup_dir, "size")) as f:
                backup_file_raw_size = sizeof_fmt(int(f.read()))
            backup_list.append("{id}: {date} {time} {tz}, {recipe_name}, {backup_file_size} ({percent_of_total} of disk) ({backup_file_raw_size} raw)"
                               .format(id=m_id, date=m_date, time=m_time, tz=m_tz, recipe_name=m_recipe_name, backup_file_size=backup_file_size, percent_of_total=percent_of_total, backup_file_raw_size=backup_file_raw_size))
    return backup_list


@pythonify_args
def backup_rm(backup_id):
    for node in os.listdir(BACKUPS_DIR):
        match = re.match(re_backup_name, node)
        if match is not None:
            m_id = match.group(1)
            if m_id == backup_id:
                backup_dir = os.path.join(BACKUPS_DIR, node)
                deleted_backup_dir = os.path.join(BACKUPS_DIR, "deleted-backup")
                subprocess.check_call("mv {backup_dir} {deleted_backup_dir}".format(backup_dir=backup_dir, deleted_backup_dir=deleted_backup_dir), shell=True)
                sync_dir(CODE_DIR)
                subprocess.check_call("rm -rf {deleted_backup_dir}".format(deleted_backup_dir=deleted_backup_dir), shell=True)
                break


@pythonify_args
def clean(datasets):
    for dataset in datasets:
        if dataset == 'state':
            log("cleaning {dataset}".format(dataset=dataset))
            for node in os.listdir(STATE_DIR):
                if node != "lost+found":
                    node_path = os.path.join(STATE_DIR, node)
                    subprocess.check_call("rm -rf {node_path}".format(node_path=node_path), shell=True)
            log("{dataset} was cleaned".format(dataset=dataset))
        elif dataset == 'code':
            log("cleaning {dataset}".format(dataset=dataset))
            pacman_db_dir = os.path.join(STATE_DIR, ".pacman", "db")
            subprocess.check_call("rm -rf {pacman_db_dir}".format(pacman_db_dir=pacman_db_dir), shell=True)
            for node in os.listdir(CODE_DIR):
                if node not in ("lost+found", ".jsc", ".pacman"):
                    node_path = os.path.join(CODE_DIR, node)
                    subprocess.check_call("rm -rf {node_path}".format(node_path=node_path), shell=True)
            log("{dataset} was cleaned".format(dataset=dataset))


@pythonify_args
def revert(backup_id):
    if not is_code_dir_clean():
        return False, "{code_dir} is not clean".format(code_dir=CODE_DIR)
    for backup_dir in os.listdir(BACKUPS_DIR):
        match = re.match(re_backup_name, backup_dir)
        if match is not None:
            m_id = match.group(1)
            if m_id == backup_id:
                backup_dir_path = os.path.join(BACKUPS_DIR, backup_dir)
                backup_archive = os.path.join(backup_dir_path, "data.tar.lzo")
                subprocess.check_call("tar -xf {backup_archive} -C /".format(backup_archive=backup_archive), shell=True)
                recipe_file_path = os.path.join(backup_dir_path, "recipe")
                if os.path.isfile(recipe_file_path):
                    subprocess.call("mv {recipe_file_path {jsc_dir}/new-recipe".format(recipe_file_path=recipe_file_path, jsc_dir=JSC_DIR), shell=True)
    return True, None


@pythonify_args
def run():
    poller = select.epoll()
    try:
        subproc = subprocess.Popen(["{code_dir}/init".format(code_dir=CODE_DIR)], stdout=subprocess.PIPE)
    except FileNotFoundError:
        return False, "No init in {code_dir}".format(code_dir=CODE_DIR)
    poller.register(subproc.stdout, select.EPOLLHUP | select.EPOLLERR | select.EPOLLIN)
    while True:
        for fd, flags in poller.poll():
            if flags == select.EPOLLIN:
                sys.stdout.write(os.read(fd))
            elif flags == select.EPOLLHUP or flags == select.EPOLLERR:
                poller.unregister(fd)
                break
        break
    return True,


@pythonify_args
def sync():
    recipe_path = os.path.join(JSC_DIR, "recipe")
    if not os.path.isdir(recipe_path):
        return True, None
    is_synced_file_path = os.path.join(recipe_path, "is-software-list-synced")
    with open(is_synced_file_path) as f:
        if int(f.read().strip()) != 0:
            return True, None
    env = env_json()
    if "software_list_sync_url" in env['ident']['container']:
        with open("{recipe_path}/software-list".format(recipe_path=RECIPE_PATH)) as f:
            software_list_content = f.read()
        try:
            request = url.Request(env["ident"]["container"]["software_list_sync_url"], data=software_list_content.encode())
            request.add_header("Authorization", "Session-Key {session_key}".format(session_key=env["ident"]["container"]["session_key"]))
            response = url.urlopen(request)
            if response.status != httplib.OK:
                return False, response.reason
            with open(is_synced_file_path, "w") as f:
                f.truncate(0)
                f.write("1")
                f.flush()
        except urllib.error.HTTPError as e:
            return False, "{}".format(e)
    return True, None


@pythonify_args
def recipe_reset():
    if os.path.exists(RECIPE_PATH):
        shutil.rmtree(RECIPE_PATH)
    # For consecutive deploys to work we need to clear every time.
    if os.path.exists(NEW_RECIPE_PATH):
        shutil.rmtree(NEW_RECIPE_PATH)
    os.mkdir(NEW_RECIPE_PATH)
    os.mkdir(NEW_RECIPE_SRC)


@pythonify_args
def read_new_recipe():
    if os.path.exists(NEW_RECIPE_SCRIPT):
        with open(NEW_RECIPE_SCRIPT) as f:
            return f.read()


@pythonify_args
def write_file(args):
    path, content = args
    with open(path, 'w') as f:
        f.write(content)


@pythonify_args
def now3339():
    return datetime.datetime.now().replace(microsecond=0).isoformat('T')


@pythonify_args
def mvtree(args):
    src, dst = args
    shutil.move(src, dst)


@pythonify_args
def rmtree(src):
    shutil.rmtree(src)


@pythonify_args
def package(args):
    output = {}
    proc = subprocess.Popen(["jumpstart", "--noconfirm", "-Sy"] + args, stdout=subprocess.PIPE)
    while True:
        line = proc.stdout.readline().decode('utf-8')
        if line == "":
            break
        words = line.split(" ")
        if words[0] == "Packages":
            # First we expect the number of packages, in a paren like (23)
            n_pkg = int(words[1].strip("()"))
            for i in range(0, n_pkg):
                # For each package we expect 2 list elements
                pkg_full = words[2 + 2 * i]
                # The package version is concated to the package name with a - between.
                pkg_parts = pkg_full.split("-")
                pkg_name = pkg_parts[0]
                # The version can and often do contain aditional '-'
                pkg_version = "-".join(pkg_parts[1:]).strip()
                if pkg_name not in args:
                    return False, "No such package"
                output[pkg_name] = pkg_version
    return True, output


@pythonify_args
def git_clone(src, dst, depth, branch, pkey):
    dst = os.path.join(JSC_DIR, "new-recipe", dst)
    depth_str = "--depth %s" % depth if depth is not None else ""
    branch_str = "--branch %s" % branch if branch is not None else ""
    git_cmd = "git clone {depth} {branch} {src} {dst}".format(depth=depth_str, branch=branch_str, src=src, dst=dst)
    if pkey is None:
        clone_cmd = git_cmd
    else:
        clone_cmd = "ssh-agent (ssh-add {pkey} && {git_cmd})".format(pkey=pkey, git_cmd=git_cmd)
    git_ret = subprocess.check_call(clone_cmd, shell=True)
    if git_ret != 0:
        return False, "git clone failed"
    with open(os.path.join(dst, ".git", "HEAD")) as head_f:
        head_ref = head_f.readline().strip()
    head_ref_parts = head_ref.split(" ")
    if len(head_ref_parts) > 1:
        # We have a non-commit ref
        if head_ref_parts[0] != "ref:":
            return False, "Invalid ref spec in %s" % src
        ref = head_ref_parts[1].strip()
        # Path to ref spec, so we can read out commit
        ref_path = os.path.join(dst, ".git", ref)
        with open(ref_path) as ref_f:
            commit = ref_f.read().strip()
        # We know all we need to know
        return True, (commit, ref)
    # Commit was in .git/HEAD
    return True, (head_ref, None)


# remoto magic, needed to execute these functions remotely
if __name__ == '__channelexec__':
    try:
        for item in channel:  # noqa
            channel.send(eval(item))  # noqa
    except KeyboardInterrupt:
        pass



import sys
import threading


class bcolors:
    HEADER = '\033[95m'
    OKBLUE = '\033[94m'
    OKGREEN = '\033[92m'
    WARNING = '\033[93m'
    WHITE = '\033[97m'
    FAIL = '\033[91m'
    ENDC = '\033[0m'
    BOLD = "\033[1m"

    def disable(self):
        self.HEADER = ''
        self.OKBLUE = ''
        self.OKGREEN = ''
        self.WARNING = ''
        self.WHITE = ''
        self.FAIL = ''
        self.ENDC = ''


print_lock = threading.Lock()


def print_locked(message, bcolor, printstr="%s%s%s\n"):
    with print_lock:
        sys.stderr.write(printstr % (bcolor, message, bcolors.ENDC))
        sys.stderr.flush()


def info(message):
    print_locked(message, bcolors.HEADER)


def white(message):
    print_locked(message, bcolors.WHITE)


def ok(message):
    print_locked(message, bcolors.OKGREEN)


def warn(message):
    print_locked(message, bcolors.WARNING)


def err(message):
    print_locked(message, bcolors.FAIL, "%sError: %s%s\n")



import os
import os.path
import sys

from setuptools import setup, find_packages

from jsc import __version__, __description__, __long_description__


def read(fname):
    path = os.path.join(os.path.dirname(__file__), fname)
    f = open(path)
    return f.read()

install_requires = []
pyversion = sys.version_info[:2]
if pyversion < (2, 7) or (3, 0) <= pyversion <= (3, 1):
    install_requires.append('argparse')


setup(
    name='jsc',
    version=__version__,
    packages=find_packages(),
    author='Jumpstarter',
    author_email='team@jumpstarter.io',
    description=__description__,
    long_description=__long_description__,
    license='MIT',
    keywords='jumpstarter js jsc deploy',
    url="https://github.com/jumpstarter-io/jsc",
    zip_safe=True,
    include_package_data=True,

    install_requires=[
        'setuptools',
        'remoto',
        'choice',
        'docopt'
        ] + install_requires,

    entry_points={
        'console_scripts': [
            'jsc = jsc.main:main',
            ]
        }
    )



import atexit
import argparse
import choice
import cmd
import distutils.version as dist_version
from docopt import docopt, DocoptExit
import json
import logging
import multiprocessing
import os
import os.path
import platform
import remoto
import remoto.process
import execnet
import shlex
import signal
import subprocess
import sys
import time
try:
    import recipe
    import logger as log
    import container
except ImportError:
    from jsc import recipe, logger as log, container
try:
    import urllib2 as url
except ImportError:
    import urllib.request as url


import pkg_resources  # part of setuptools
__version__ = pkg_resources.require("jsc")[0].version


DEFAULT_SSH_HOST = "ssh.jumpstarter.io"
DEFAULT_SSH_PORT = "22"

PYPI_JSON = "https://pypi.python.org/pypi/jsc/json"

MINUTE_S = 60
MIN_TIME_BETWEEN_UPDATES = MINUTE_S * 10


CODE_DIR = "/app/code"
STATE_DIR = "/app/state"
JSC_DIR = os.path.join(CODE_DIR, ".jsc")
NEW_RECIPE_PATH = os.path.join(JSC_DIR, "new-recipe")
RECIPE_PATH = os.path.join(JSC_DIR, "recipe")
NEW_RECIPE_SRC = os.path.join(NEW_RECIPE_PATH, "src")
NEW_RECIPE_SCRIPT = os.path.join(NEW_RECIPE_SRC, "Jumpstart-Recipe")


def fail(message):
    log.err(message)
    os._exit(1)


def stop(message):
    log.warn(message)
    os._exit(0)


def rsync(conn, src, dst):
    class JSRSync(execnet.RSync):
        def _report_send_file(self, gateway, modified_rel_path):
            log.info("syncing file: %s" % modified_rel_path)
    sync = JSRSync(src)
    sync.add_target(conn.gateway, dst)
    sync.send()


def docopt_cmd(func):
    """
    This decorator is used to simplify the try/except block and pass the result
    of the docopt parsing to the called action.
    """
    def fn(self, arg):
        try:
            opt = docopt(fn.__doc__, shlex.split(arg))
        except DocoptExit as e:
            # The DocoptExit is thrown when the args do not match.
            # We print a message to the user and the usage block.
            log.err('Invalid Command!')
            log.err(e)
            return
        except SystemExit:
            # The SystemExit exception prints the usage for --help
            # We do not need to do the print here.
            return
        return func(self, opt)
    fn.__name__ = func.__name__
    fn.__doc__ = func.__doc__
    fn.__dict__.update(func.__dict__)
    return fn


class Console(cmd.Cmd):
    def __init__(self, ssh_username, remoto_exec, ssh_conn_str):
        cmd.Cmd.__init__(self)
        self.prompt = "{ssh_username}> ".format(ssh_username=ssh_username)
        self._remoto_exec = remoto_exec
        self._ssh_conn_str = ssh_conn_str

    @docopt_cmd
    def do_backup(self, args):
        """
        Usage:
          backup [ls]
          backup new
          backup du
          backup rm <id>

        Managing backups.

        Arguments:
          ls        List backups.
          new       Create a new backup.
          du        Like ls, but also shows size and disk space usage.
          rm        Removes backup with supplied id.
        """
        if args['new']:
            self._remoto_exec.backup_new()
            log.ok("Backup done!")
        elif args['du']:
            for backup in self._remoto_exec.backup_du():
                log.info(backup)
        elif args['rm']:
            backup_id = args['<id>']
            self._remoto_exec.backup_rm(backup_id)
            log.info("Removal of backup [{backup_id}] done!".format(backup_id=backup_id))
        else:
            for backup in self._remoto_exec.backup_ls():
                log.info(backup)

    @docopt_cmd
    def do_clean(self, args):
        """
        Usage:
          clean [--code|--state|--all]

        Cleans all user added directories and files in the given dataset.

        Options:
          --code            Cleans /app/code (default).
          --state           Cleans /app/state.
          --all             Cleans both /app/code and /app/state.
        """
        if args['--all']:
            datasets = ['state', 'code']
        else:
            datasets = [item[2:] for item in args if args[item] is True]
        if len(datasets) == 0:
            datasets = ['code']
        self._remoto_exec.clean(datasets)

    @docopt_cmd
    def do_clone(self, args):
        """
        Usage:
          clone <id> [--pkey=<pkey>] [--code|--state]

          NOT IMPLEMENTED
        """
        # Low priotiry. Do later.
        pass

    @docopt_cmd
    def do_deploy(self, args):
        """
        Usage:
          deploy [--dev] <path>

        Deployes a user writter recipe.

        Arguments:
          <path>        A path to the local recipe or a git repo which contains a recipe.

        Options:
          --dev         Uses git clone instead of git archive to keep .git.
                        This should NOT be done on an assembly that are going to be released.
        """
        if not self._remoto_exec.is_code_dir_clean():
            log.err("deploy requires clean /app/code")
            return
        path = args["<path>"]
        is_dev = "1" if args['--dev'] else "0"
        # From spec:
        # 1. Deleting .jsc/recipe and initializing .jsc/new-recipe. This folder
        #    should be deleted in init if it is found as it signifies a
        #    non-completed deploy.
        self._remoto_exec.recipe_reset()
        # 2. The full recipe is copied over into .jsc/new-recipe/src. If the
        #    source is a git repo the .git folder should not be included.
        # Depending on what path is we need to do different things
        # 1 Path to a local file it is a recipe, copy and execute.
        # 2 Path to a local dir it contains a recipe and resource for the recipe.
        #   Copy over all and execute.
        # 3 URL to a git repository, clone in and execute recipe.
        # 4 "GitHub ID" Clone and execute.
        if os.path.isfile(path):
            self._remoto_exec.rsync(path, NEW_RECIPE_SCRIPT)
        elif os.path.isdir(path):
            self._remoto_exec.rsync(path, NEW_RECIPE_SRC)
        else:
            path_parts = path.split(":")
            if path_parts[0] == "github":
                # github repo id format.
                repo_url = "git://github/" + path_parts[0:].join(":") + ".git"
            else:
                repo_url = path
            self._remoto_exec.git_clone(repo_url, "src", 1, None, None)
            self._remoto_exec.rmtree(os.path.join(NEW_RECIPE_SRC, ".git"))
        # 3. A disk sync is performed on /app/code.
        self._remoto_exec.sync_dir('/')
        # 4. Syncing the jumpstart repo so it's up to date.
        # Not needed, jumpstart -Sy is a better solution
        # 5. The recipe (.jsc/new-recipe/src/Jumpstart-Recipe) is executed by
        #    interpretation in jsc.
        recipe_script = self._remoto_exec.read_new_recipe()
        rec_o = recipe.run(self._remoto_exec, recipe_script, is_dev)
        if rec_o is None:
            log.err("Deploy unsuccessful")
            return
        # 6. A disk sync is performed on /app/code.
        self._remoto_exec.sync_dir(CODE_DIR)
        # 7. The software list is exported as JSON to .jsc/new-recipe/software-list.
        software_list = json.dumps(rec_o["software_list"], sort_keys=True, indent=4, separators=(',', ': '))
        self._remoto_exec.write_file([os.path.join(NEW_RECIPE_PATH, "software-list"), software_list])
        # 8. The file .jsc/new-recipe/is-dev is created with the content 1 when
        #    --dev is specified, otherwise 0.
        self._remoto_exec.write_file([os.path.join(NEW_RECIPE_PATH, "is-dev"), is_dev])
        # 9. The file .jsc/new-recipe/is-software-list-synced is created with
        #    the content 0.
        is_software_list_path = os.path.join(NEW_RECIPE_PATH, "is-software-list-synced")
        self._remoto_exec.write_file([is_software_list_path, "0"])
        # 10. The file .jsc/new-recipe/deploy-time is created with the current
        #     time in rfc 3339 format with zero precision.
        remote_time = self._remoto_exec.now3339()
        self._remoto_exec.write_file([os.path.join(NEW_RECIPE_PATH, "deploy-time"), remote_time])
        # 11. A disk sync is performed on /app/code.
        self._remoto_exec.sync_dir(CODE_DIR)
        # 12. Moving .jsc/new-recipe to .jsc/recipe.
        self._remoto_exec.mvtree([NEW_RECIPE_PATH, RECIPE_PATH])
        # 13. A disk sync is performed on /app/code.
        self._remoto_exec.sync_dir(CODE_DIR)
        # 14. A software list sync is performed.
        success, msg = self._remoto_exec.sync()
        if not success:
            log.warn("Sync failed because of: [{msg}]. Try running clean manually.".format(msg=msg))
        # 15. Informing the user that the deploy was succesful.
        log.info("Deploy successful!")

    @docopt_cmd
    def do_env(self, args):
        """
        Usage:
          env

        Dumps /app/env.json to console.
        """
        env_json = self._remoto_exec.env_json()
        log.info(json.dumps(env_json, sort_keys=True, indent=4, separators=(',', ': ')))

    @docopt_cmd
    def do_revert(self, args):
        """
        Usage:
          revert <id>

        Reverts a backup. This will destroy any changes you've made since backup.

        Arguments:
          <id>        Id of the backup to restore.
        """
        backup_id = args['<id>']
        success, msg = self._remoto_exec.revert(backup_id)
        if not success:
            log.err(msg)
            return
        log.info("revert of backup [{backup_id}] done!".format(backup_id=backup_id))

    @docopt_cmd
    def do_run(self, args):
        """
        Usage:
          run

        Runs /app/code/init.
        """
        try:
            success, msg = self._remoto_exec.run()
            if not success:
                log.err(msg)
        except KeyboardInterrupt:
            # give the prompt a fresh new line instead of ^C
            print("")

    @docopt_cmd
    def do_ssh(self, args):
        """
        Usage:
          ssh

        Opens a new terminal and starts an SSH shell to this assembly.
        """
        if sys.platform.startswith("linux"):
            cmd_open = "xdg-open"
        elif sys.platform.startswith("darwin"):
            cmd_open = "open"
        elif sys.platform.startswith("win32"):
            cmd_open = "start"
        else:
            log.err("Cannot start ssh shell since your platform [{platform}] is not supported.".format(platform=platform))
            return
        try:
            subprocess.check_call([cmd_open, "ssh://{uri}".format(uri=self._ssh_conn_str)])
        except subprocess.CalledProcessError:
            pass

    @docopt_cmd
    def do_sync(self, args):
        """
        Usage:
          sync

        Syncs software list so it becomes visable in the developer panel.
        """
        success, msg = self._remoto_exec.sync()
        if not success:
            log.warn("Sync failed because of: [{msg}]. Please contact support if problem remains.".format(msg=msg))
        else:
            log.ok("Sync successful!")

    @docopt_cmd
    def do_status(self, args):
        """
        Usage:
          status [-v]

        Shows status about the container.
        Like disk usage, recipe deployed, backups and more.

        Options:
          -v         Be more verbose and show what jumpstart-packages are installed and
                     which git repos are checked out.
        """
        verbose = args["-v"]
        self._remoto_exec.status_short(__version__, verbose)

    def do_hist(self, args):
        """Print a list of commands that have been entered"""
        log.info(self._hist)

    def do_exit(self, args):
        """Exits from the console"""
        log.info(args)
        return -1

    def do_quit(self, args):
        self.do_exit("")

    def do_EOF(self, args):
        """Exit on system end of file character"""
        return self.do_exit(args)

    def preloop(self):
        """Initialization before prompting user for commands.
           Despite the claims in the Cmd documentaion, Cmd.preloop() is not a stub.
        """
        cmd.Cmd.preloop(self)   ## sets up command completion
        self._hist    = []      ## No history yet
        self._locals  = {}      ## Initialize execution namespace for user
        self._globals = {}

    def postloop(self):
        """Take care of any unfinished business.
           Despite the claims in the Cmd documentaion, Cmd.postloop() is not a stub.
        """
        cmd.Cmd.postloop(self)   ## Clean up command completion
        log.info("Goodbye!")

    def precmd(self, line):
        """ This method is called after the line has been input but before
            it has been interpreted. If you want to modifdy the input line
            before execution (for example, variable substitution) do it here.
        """
        self._hist += [line.strip()]
        return line

    def postcmd(self, stop, line):
        """If you want to stop the console, return something that evaluates to true.
           If you want to do some post command processing, do it here.
        """
        return stop

    def emptyline(self):
        """Do nothing on empty input line"""
        pass

    def default(self, line):
        """Called on an input line when the command prefix is not recognized.
           In that case we execute the line as Python code.
        """
        log.info("unknown command: [{line}]".format(line=line))


def touch_dir(directory):
    if not os.path.exists(directory):
        try:
            os.makedirs(directory)
        except FileExistsError:
            pass


def update_self():
    # TODO: implement
    state_dir = os.path.expanduser("~/.jsc")
    touch_dir(state_dir)
    last_update_file = os.path.join(state_dir, "last_update")
    last_update_time = 0
    if os.path.exists(last_update_file):
        with open(last_update_file) as f:
            last_update_time = int(f.read().strip())
    epoch_time = int(time.time())
    if (epoch_time - last_update_time) > MIN_TIME_BETWEEN_UPDATES:
        log.info("Checking for updates of jsc")
        try:
            response = url.urlopen(PYPI_JSON)
            package_json = json.loads(response.read().decode())
            if dist_version.StrictVersion(package_json['info']['version']) > dist_version.StrictVersion(__version__):
                stop("There's a new version available, update with '# pip install -U jsc'")
            else:
                log.info("You're running the latest version of jsc")
            with open(last_update_file, "w") as f:
                f.truncate(0)
                f.write(str(epoch_time))
        except (url.URLError, ValueError):
            log.warn("Could not check for updates, try '# pip install -U jsc'")


def parse_args_main(args):
    parser = argparse.ArgumentParser(description='Jumpstarter Console')
    parser.add_argument('-V', '--version', action='version', version='{version}'.format(version=__version__))
    parser.add_argument('-H', '--host', action='store', default=DEFAULT_SSH_HOST, help="Hostname of ssh endpoint. Default: ssh.jumpstarter.io")
    parser.add_argument('-p', '--port', action='store', default=DEFAULT_SSH_PORT, help="Port of ssh endpoint. Default: 22")
    parser.add_argument('--no-update', action='store_true', help="Do not check for updates of jsc")
    parser.add_argument('ssh_username', help="SSH username")
    return parser.parse_args(args)


class RemotoProcessCom():
    def __init__(self, q_in, q_out):
        self._q_in = q_in
        self._q_out = q_out

    def _caller(self, func_identifier):
        def func(*args, **kwargs):
            self._q_in.put((func_identifier, args, kwargs))
            return self._q_out.get()
        return func

    def __getattr__(self, item):
        if hasattr(container, item):
            return self._caller(item)
        raise AttributeError("'{class_name}' object has no attribute '{item}'".format(class_name=self.__class__.__name__, item=item))

    def rsync(self, src, dst):
        self._q_in.put(('rsync', [src, dst], {}))
        return self._q_out.get()


def remoto_process(ssh_conn_str, q_in, q_out):
    signal.signal(signal.SIGINT, signal.SIG_IGN)
    try:
        conn = remoto.Connection(ssh_conn_str, logger=logging.getLogger(ssh_conn_str), threads=0)
    except KeyboardInterrupt:
        q_out.put((False, "connection canceled by user"))
        return
    except execnet.gateway_bootstrap.HostNotFound:
        q_out.put((False, "could not connect"))
        return
    q_out.put((True, None))
    conn.import_module(container)
    while True:
        func_identifier, args, kwargs = q_in.get()
        if func_identifier == "rsync":
            q_out.put(rsync(conn, *args, **kwargs))
            continue
        func = getattr(conn.remote_module, func_identifier)
        q_out.put(func("remoto", [args, kwargs]))


def kill_process(process):
    process.terminate()


def process_signal(process):
    def signal_handler(signum, frame):
        if signum == signal.SIGINT:
            kill_process(process)
            os._exit(1)
    return signal_handler


def main(args=None):
    signal.signal(signal.SIGINT, signal.SIG_IGN)
    try:
        parsed = parse_args_main(args)
        if not parsed.no_update:
            update_self()
        # WARNING: port does is not supported by remoto atm
        ssh_username = parsed.ssh_username
        ssh_conn_str = "{id}@{host}".format(id=ssh_username, host=parsed.host, port=parsed.port)
        # Init
        remoto_q_in = multiprocessing.Queue()
        remoto_q_out = multiprocessing.Queue()
        remoto_t = multiprocessing.Process(target=remoto_process, args=(ssh_conn_str, remoto_q_in, remoto_q_out))
        remoto_t.start()
        signal.signal(signal.SIGINT, process_signal(remoto_t))
        # kill the remoto process if main process dies
        atexit.register(kill_process, remoto_t)
        try:
            success, msg = remoto_q_out.get()
        except KeyboardInterrupt:
            success, msg = remoto_q_out.get()
        if not success:
            fail(msg)
        remoto_exec = RemotoProcessCom(remoto_q_in, remoto_q_out)
        is_assembly, msg = remoto_exec.assert_is_assembly()
        if not is_assembly:
            fail(msg)
        success, _ = remoto_exec.check_init()
        if not success:
            confirm = choice.Binary('Assembly is not initialized, would you like to do it now?', False).ask()
            if not confirm:
                stop("You choose not to init the assembly, exiting...")
        remoto_exec.init()
        lock_content_json = {
            "hostname": platform.node(),
            "unix_epoch": int(time.time()),
        }
        lock_content = json.dumps(lock_content_json)
        success, msg = remoto_exec.lock_session(lock_content)
        if not success:
            fail(msg)
        # TODO: clean incomplete backup stuff
        success, msg = remoto_exec.sync()
        if not success:
            log.warn("Sync failed because of: [{msg}]. Try running clean manually.".format(msg=msg))
        if not success:
            log.err(msg)
        # Print status on login
        remoto_exec.status_short(__version__, False)
        # Start console prompt
        console = Console(ssh_username, remoto_exec, ssh_conn_str)
        console.cmdloop("Welcome to jsc!")
    except KeyboardInterrupt:
        pass
    kill_process(remoto_t)
    os._exit(1)


if __name__ == '__main__':
    main()
    pass
